/*
 * async.h - Utilities for async operations
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#ifndef INCLUDED_ASYNC_H
#define INCLUDED_ASYNC_H

#include <stdbool.h>
#include <sys/types.h>

typedef enum {
    ASYNC_OPERATION_DONE,
    ASYNC_OPERATION_FAILED,
    ASYNC_OPERATION_WORKING
} AsyncOperationStatus;

typedef struct _AsyncOperation AsyncOperation;

// true on success, false otherwise
typedef bool(*AsyncOperationPollFunc)(AsyncOperation *op);
typedef void(*AsyncOperationCleanupFunc)(AsyncOperation *op);

struct _AsyncOperation {
    AsyncOperationPollFunc poll_func;
    AsyncOperationCleanupFunc cleanup_func;
    int status;
    pid_t pid;
    bool done;
    void *data1;
    void *data2;
};
#define async_operation_failed(o) (WEXITSTATUS((o)->status) != 0)

/*
 * Start a new async operation using the executable PATH and the arguments
 * ARGS. The first element of ARGS should be the name of the new process, by
 * convention. DATA1 and DATA2 are arbitrary pointers for storing data needed by
 * CLEANUP or POLL.
 * Return: NULL on error, the operation otherwise
 */
AsyncOperation *async_operation_start(AsyncOperationPollFunc poll,
                                      AsyncOperationCleanupFunc cleanup,
                                      void *data1, void *data2,
                                      const char *path, char * const *args);

/*
 * Create a new async operation from an already running process with PID.
 */
AsyncOperation *async_operation_take(AsyncOperationPollFunc poll,
                                     AsyncOperationCleanupFunc cleanup,
                                     void *data1, void *data2, pid_t pid);

/*
 * Poll and return the status of OP. If HANG, wait until OP is done.
 */
AsyncOperationStatus async_operation_poll(AsyncOperation *op, bool hang);

/*
 * Kill OP, by FORCE if necessary. This does NOT free the process, you need to
 * do that yourself.
 */
void async_operation_cancel(AsyncOperation *op, bool force);

/*
 * Free OP by calling its cleanup function. This will cancel it if it's still
 * running!
 */
void async_operation_free(AsyncOperation *op);

#endif
