/*
 * setdatescreen.c - Screen for setting the date and time
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#include "setdatescreen.h"

#include <err.h>
#include <unistd.h>
#include <sys/time.h>

static void set_date_reset(SetDateScreen *screen,
                           SensorState *state) {
    date_sel_reset(&screen->ds);
    time_sel_reset(&screen->ts);
    screen->need_redraw = true;
    screen->stage = SET_DATE_SELECT_DATE;
    lcd_display_control(state->lcd, LCD_CURSOR_NO_BLINK, LCD_CURSOR_OFF,
                        LCD_DISPLAY_ON);
}

static bool set_date_screen_select_date(SetDateScreen *screen,
                                        SensorState *state) {
    if (state->sel_down || state->back_down || state->up_down ||
        state->down_down) {
        screen->need_redraw = true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        DateSelectionState dss = date_sel_dispatch(&screen->ds, state, "New Date:");
        switch (dss) {
        case DATE_SEL_CONTINUE:
            return false; // ignore
        case DATE_SEL_BACK:
            return true;
        case DATE_SEL_DONE:
            screen->stage = SET_DATE_SELECT_TIME;
            break;
        case DATE_SEL_ERROR:
            screen->stage = SET_DATE_FAIL;
            break;
        }
        screen->need_redraw = true;
    }
    return false;
}

static bool set_date_and_time(DateSelection *ds, TimeSelection *ts) {
    int64_t utc_time = timedate_sel_to_uts_timestamp(ds, ts);
    struct timeval new_time = {
        .tv_sec = utc_time,
        .tv_usec = 0, // can't set, user not fast enough (lol)
    };
    if (settimeofday(&new_time, NULL) < 0) {
        warn("settimeofday(2) failed");
        return false;
    }
    LOG_VERBOSE("Set system date and time to %04d/%02d/%02d %02d:%02d:%02d\n",
                ds->year, ds->month, ds->day, ts->hour, ts->minute, ts->second);
    return true;
}

static void set_date_screen_select_time(SetDateScreen *screen,
                                        SensorState *state) {
    if (state->sel_down || state->back_down || state->up_down ||
        state->down_down) {
        screen->need_redraw = true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        TimeSelState tss = time_sel_dispatch(&screen->ts, state, "New Time");
        switch (tss) {
        case TIME_SEL_CONTINUE:
            return; // ignore
        case TIME_SEL_BACK:
            --screen->stage;
            break;
        case TIME_SEL_DONE:
            if (set_date_and_time(&screen->ds, &screen->ts)) {
                screen->stage = SET_DATE_SUCCESS;
            } else {
                screen->stage = SET_DATE_FAIL;
            }
            break;
        case TIME_SEL_ERROR:
            screen->stage = SET_DATE_FAIL;
            break;
        }
        screen->need_redraw = true;
    }
}

static bool set_date_screen_message(SetDateScreen *screen,
                                    SensorState *state,
                                    const char *msg) {
    if (state->up_down || state->down_down || state->sel_down ||
        state->back_down) {
        return true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, msg);
        lcd_display_control(state->lcd, LCD_CURSOR_NO_BLINK, LCD_CURSOR_OFF,
                            LCD_DISPLAY_ON);
    }
    return false;
}

static bool set_date_screen_dispatch(SetDateScreen *screen,
                                     SensorState *state) {
    if (state->force_draw) {
        screen->need_redraw = true;
    }
    switch (screen->stage) {
    case SET_DATE_SELECT_DATE:
        if (set_date_screen_select_date(screen, state)) {
            set_date_reset(screen, state);
            return true;
        }
        break;
    case SET_DATE_SELECT_TIME:
        set_date_screen_select_time(screen, state);
        break;
    case SET_DATE_FAIL:
        if (set_date_screen_message(screen, state, "Fail!")) {
            set_date_reset(screen, state);
            return true;
        }
        break;
    case SET_DATE_SUCCESS:
        if (set_date_screen_message(screen, state, "Success!")) {
            set_date_reset(screen, state);
            return true;
        }
        break;
    }
    return false;
}

SetDateScreen *set_date_screen_new() {
    SetDateScreen *s = malloc_checked(sizeof(SetDateScreen));
    screen_init(&s->parent, "Set date/time",
                (ScreenDispatchFunc) set_date_screen_dispatch,
                (ScreenCleanupFunc) free);
    date_sel_init(&s->ds, PERIOD_DAY);
    date_sel_set_clamp_time(&s->ds, false);
    time_sel_init(&s->ts, NULL, true);
    s->need_redraw = true;
    s->stage = SET_DATE_SELECT_DATE;
    return s;
}
