/*
 * settzscreen.c - Screen for setting the timezone
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#include "settzscreen.h"

#include <spawn.h>
#include <err.h>
#include <sys/wait.h>

extern char **environ;

static void set_tz_screen_reset(SetTZScreen *screen, SensorState *state) {
    screen->has_dst = false;
    screen->stage = SET_TZ_OFFSET;
    screen->off_sel_stage = TZ_OFFSET_SEL_HOUR;
    screen->bound_sel_stage = DST_BOUND_SEL_TYPE;
    tz_offset_reset(&screen->offset);
    tz_offset_reset(&screen->dst_offset);
    screen->did_set = false;
    if (state) {
        lcd_display_control(state->lcd, LCD_CURSOR_NO_BLINK, LCD_CURSOR_OFF,
                            LCD_DISPLAY_ON);
    }
}

static void set_tz_offset_normalize(TZOffset *offset, bool allow_negative) {
    if (offset->second < 0) {
        offset->second = 59;
        --offset->minute;
    } else if (offset->second > 59) {
        offset->second = 0;
        ++offset->minute;
    }
    if (offset->minute < 0) {
        offset->minute = 59;
        --offset->hour;
    } else if (offset->minute > 59) {
        offset->minute = 0;
        ++offset->hour;
    }
    if (allow_negative) {
        if (offset->hour < -24) {
            offset->hour = 24;
        } else if (offset->hour > 24) {
            offset->hour = -24;
        }
    } else {
        if (offset->hour < 0) {
            offset->hour = 24;
        } else if (offset->hour > 24) {
            offset->hour = 0;
        }
    }
}

static SetTZScreenAction set_tz_dispatch_select_offset(TZOffsetSelectStage *stage,
                                                       SensorState *state,
                                                       TZOffset *target,
                                                       const char *label,
                                                       bool force_redraw,
                                                       bool allow_negative) {
    if (state->back_down) {
        if (*stage == TZ_OFFSET_SEL_HOUR) {
            *stage = TZ_OFFSET_SEL_SECOND;
            return SET_TZ_BACK;
        }
        --*stage;
        force_redraw = true;
    }
    if (state->up_down || state->down_down) {
        force_redraw = true;
        int change = state->up_down - state->down_down;
        switch (*stage) {
        case TZ_OFFSET_SEL_HOUR:
            target->hour += change;
            break;
        case TZ_OFFSET_SEL_MINUTE:
            target->minute += change;
            break;
        case TZ_OFFSET_SEL_SECOND:
            target->second += change;
            break;
        }
        set_tz_offset_normalize(target, allow_negative);
    }
    if (state->sel_down) {
        if (*stage == TZ_OFFSET_SEL_SECOND) {
            *stage = TZ_OFFSET_SEL_HOUR;
            return SET_TZ_NEXT;
        }
        ++*stage;
        force_redraw = true;
    }
    if (force_redraw) {
        const char *format;
        int cursor_pos;
        switch (*stage) {
        case TZ_OFFSET_SEL_HOUR:
            format = ">%0*d:%02d:%02d";
            cursor_pos = 0;
            break;
        case TZ_OFFSET_SEL_MINUTE:
            format = "%0*d:>%02d:%02d";
            cursor_pos = target->hour < 0 ? 4 : 3;
            break;
        case TZ_OFFSET_SEL_SECOND:
            format = "%0*d:%02d:>%02d";
            cursor_pos = target->hour < 0 ? 7 : 6;
            break;
        default:
            format = "";
            cursor_pos = 0;
            break;
        }
        char time_str[17];
        snprintf(time_str, 17, format, target->hour < 0 ? 3 : 2, target->hour,
                 target->minute, target->second);
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, label);
        lcd_move_to(state->lcd, 1, 0);
        lcd_write_string(state->lcd, time_str);
        lcd_move_to(state->lcd, 1, cursor_pos);
        lcd_display_control(state->lcd, LCD_CURSOR_BLINK, LCD_CURSOR_ON,
                            LCD_DISPLAY_ON);
    }
    return SET_TZ_CONTINUE;
}

static void set_tz_prompt_has_dst(SetTZScreen *screen,
                                  SensorState *state) {
    if (state->back_down) {
        --screen->stage;
        screen->need_redraw = true;
        screen->off_sel_stage = TZ_OFFSET_SEL_SECOND;
        return;
    }
    if (state->sel_down) {
        if (screen->has_dst) {
            screen->stage = SET_TZ_DST_OFFSET;
        } else {
            screen->stage = SET_TZ_FINISH;
        }
        screen->need_redraw = true;
        screen->off_sel_stage = TZ_OFFSET_SEL_HOUR;
        return;
    }
    if (state->up_down || state->down_down) {
        screen->has_dst = abs((((int) screen->has_dst) + state->up_down -
                               state->down_down) % 2);
        screen->need_redraw = true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, "Has DST?");
        lcd_move_to(state->lcd, 1, 0);
        lcd_write_string(state->lcd, ">No >Yes");
        lcd_move_to(state->lcd, 1, screen->has_dst ? 4 : 0);
    } 
}

static const char *DST_BOUND_TYPE_LABELS[] = {
    "Exact Julian",
    "Leap Julian",
    "Month Day Week"
};
static SetTZScreenAction set_tz_bound_select_type(SetTZScreen *screen,
                                                  SensorState *state,
                                                  DSTBound *target,
                                                  const char *label) {
    if (state->back_down) {
        screen->off_sel_stage = TZ_OFFSET_SEL_SECOND;
        screen->bound_sel_stage = DST_BOUND_SEL_TIME;
        return SET_TZ_BACK;
    }
    if (state->sel_down) {
        switch (target->type) {
        case DST_BOUND_EXACT_JULIAN:
            screen->bound_sel_stage = DST_BOUND_SEL_JULIAN;
            target->julian_day = 1;
            break;
        case DST_BOUND_LEAP_JULAIN:
            screen->bound_sel_stage = DST_BOUND_SEL_JULIAN;
            target->julian_day = 0;
            break;
        case DST_BOUND_MND:
            screen->bound_sel_stage = DST_BOUND_SEL_MONTH;
            target->month = 1;
            target->week = 1;
            target->day = 0;
            break;
        }
        target->time.hour = 2;
        target->time.minute = 0;
        target->time.second = 0;
        screen->julian_didgit = 3;
        screen->need_redraw = true;
        return SET_TZ_CONTINUE;
    }
    if (state->up_down || state->down_down) {
        int new_type = (((int) target->type) + state->up_down - state->down_down) % 3;
        if (new_type < 0) {
            target->type = 2;
        } else {
            target->type = new_type;
        }
        screen->need_redraw = true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, label);
        lcd_write_string(state->lcd, " Type:");
        lcd_move_to(state->lcd, 1, 0);
        lcd_write_char(state->lcd, '>');
        lcd_write_string(state->lcd, DST_BOUND_TYPE_LABELS[target->type]);
        lcd_move_to(state->lcd, 1, 0);
    }
    return SET_TZ_CONTINUE;
}

static const char *MONTH_ABBREVS[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov",
    "Dec",
};
static const char *WEEKDAY_ABBREVS[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",
};
static SetTZScreenAction set_tz_bound_select_mnd(SetTZScreen *screen,
                                                 SensorState *state,
                                                 DSTBound *target,
                                                 const char *label) {
    if (state->back_down) {
        --screen->bound_sel_stage;
        screen->need_redraw = true;
        return SET_TZ_CONTINUE;
    }
    if (state->sel_down) {
        screen->need_redraw = true;
        if (screen->bound_sel_stage == DST_BOUND_SEL_DAY) {
            screen->off_sel_stage = TZ_OFFSET_SEL_HOUR;
            screen->bound_sel_stage = DST_BOUND_SEL_TIME;
            return SET_TZ_CONTINUE;
        }
        ++screen->bound_sel_stage;
    }
    if (state->up_down || state->down_down) {
        int change = state->up_down - state->down_down;
        switch (screen->bound_sel_stage) {
        case DST_BOUND_SEL_MONTH:
            target->month += change;
            if (target->month < 1) {
                target->month = 1;
            } else if (target->month > 12) {
                target->month = 12;
            }
            break;
        case DST_BOUND_SEL_WEEK:
            target->week += change;
            if (target->week < 1) {
                target->week = 1;
            } else if (target->week > 5) {
                target->week = 5;
            }
            break;
        case DST_BOUND_SEL_DAY:
            target->day += change;
            if (target->day < 0) {
                target->day = 0;
            } else if (target->day > 6) {
                target->day = 6;
            }
            break;
        default:
            break; // should not happen
        }
        screen->need_redraw = true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        const char *format;
        int cursor_pos;
        switch(screen->bound_sel_stage) {
        case DST_BOUND_SEL_MONTH:
            format = ">%s-%d-%s";
            cursor_pos = 0;
            break;
        case DST_BOUND_SEL_WEEK:
            format = "%s->%d-%s";
            cursor_pos = 4;
            break;
        case DST_BOUND_SEL_DAY:
            format = "%s-%d->%s";
            cursor_pos = 7;
            break;
        default:
            // should not happen
            format = "";
            cursor_pos = 0;
            break;
        }
        char sel_str[17];
        snprintf(sel_str, 17, format, MONTH_ABBREVS[target->month - 1],
                 target->week, WEEKDAY_ABBREVS[target->day]);
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, "Mon-Week-Day");
        lcd_move_to(state->lcd, 1, 0);
        lcd_write_string(state->lcd, sel_str);
        lcd_move_to(state->lcd, 1, cursor_pos);
    }
    return SET_TZ_CONTINUE;
}

static int set_tz_normalize_julian(int julian, bool allow_zero) {
    if (julian <= 0) {
        return allow_zero ? 0 : 1;
    } else if (julian > 365) {
        return 365;
    }
    return julian;
}

static SetTZScreenAction set_tz_dispatch_select_julian(SetTZScreen *screen,
                                                       SensorState *state,
                                                       DSTBound *target,
                                                       const char *label) {
    if (state->back_down) {
        screen->need_redraw = true;
        if (screen->julian_didgit == 3) {
            screen->bound_sel_stage = DST_BOUND_SEL_TYPE;
            return SET_TZ_CONTINUE;
        }
        ++screen->julian_didgit;
    }
    if (state->sel_down) {
        screen->need_redraw = true;
        if (screen->julian_didgit == 1) {
            screen->off_sel_stage = TZ_OFFSET_SEL_HOUR;
            screen->bound_sel_stage = DST_BOUND_SEL_TIME;
            return SET_TZ_CONTINUE;
        }
        --screen->julian_didgit;
    }
    if (state->up_down || state->down_down) {
        screen->need_redraw = true;
        int change = state->up_down - state->down_down;
        int scale = 0;
        switch (screen->julian_didgit) {
        case 1: // ones
            scale = 1;
            break;
        case 2: // tens
            scale = 10;
            break;
        case 3: // hundreds
            scale = 100;
            break;
        }
        bool allow_zero = screen->off_sel_stage == DST_BOUND_LEAP_JULAIN;
        int new_day = target->julian_day + scale * change;
        target->julian_day = set_tz_normalize_julian(new_day, allow_zero);
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        char jul_str[4];
        snprintf(jul_str, 4, "%03d", target->julian_day);
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, "Julian Day");
        lcd_move_to(state->lcd, 1, 0);
        lcd_write_string(state->lcd, jul_str);
        lcd_move_to(state->lcd, 1, 3 - screen->julian_didgit);
    }
    return SET_TZ_CONTINUE;
}

static SetTZScreenAction set_tz_dispatch_select_bound(SetTZScreen *screen,
                                                      SensorState *state,
                                                      DSTBound *target,
                                                      const char *label) {
    SetTZScreenAction action;
    switch (screen->bound_sel_stage) {
    case DST_BOUND_SEL_TYPE:
        return set_tz_bound_select_type(screen, state, target, label);
    case DST_BOUND_SEL_MONTH:
    case DST_BOUND_SEL_WEEK:
    case DST_BOUND_SEL_DAY:
        return set_tz_bound_select_mnd(screen, state, target, label);
    case DST_BOUND_SEL_JULIAN:
        return set_tz_dispatch_select_julian(screen, state, target, label);
    case DST_BOUND_SEL_TIME:
        action = set_tz_dispatch_select_offset(&screen->off_sel_stage, state,
                                               &target->time, "Effect Time:",
                                               screen->need_redraw, false);
        switch (action) {
        case SET_TZ_NEXT:
            screen->bound_sel_stage = DST_BOUND_SEL_TYPE;
        case SET_TZ_CONTINUE:
            screen->need_redraw = false;
            return action;
        case SET_TZ_BACK:
            if (target->type == DST_BOUND_MND) {
                screen->bound_sel_stage = DST_BOUND_SEL_DAY;
            } else {
                screen->julian_didgit = 1;
                screen->bound_sel_stage = DST_BOUND_SEL_JULIAN;
            }
            screen->need_redraw = true;
            return SET_TZ_CONTINUE;
        }
    default:
        return SET_TZ_BACK;
    }
}

static void set_tz_write_dst_bound(FILE *tz, DSTBound *bound) {
    fputc(',', tz);
    switch (bound->type) {
    case DST_BOUND_EXACT_JULIAN:
        fprintf(tz, "J%d", bound->julian_day);
        break;
    case DST_BOUND_LEAP_JULAIN:
        fprintf(tz, "%d", bound->julian_day);
        break;
    case DST_BOUND_MND:
        fprintf(tz, "M%d.%d.%d", bound->month, bound->week, bound->day);
        break;
    }
    fprintf(tz, "/%02d:%02d:%02d", bound->time.hour, bound->time.minute,
            bound->time.second);
}

static bool set_tz_screen_finish(SetTZScreen *screen,
                                 SensorState *state) {
    if (!screen->did_set) {
        screen->did_set =  true;
        size_t len;
        char *buf;
        FILE *tz = open_memstream(&buf, &len);
        TZOffset *off = &screen->offset;
        fprintf(tz, "timezone=CTZ%02d:%02d:%02d", off->hour, off->minute, off->second);
        if (screen->has_dst) {
            TZOffset *dst_off = &screen->dst_offset;
            fprintf(tz, "CDST%02d:%02d:%02d", dst_off->hour, dst_off->minute,
                    dst_off->second);
            set_tz_write_dst_bound(tz, &screen->start_bound);
            set_tz_write_dst_bound(tz, &screen->end_bound);
        }
        fclose(tz);
        char *const args[] = {
            "sysrc",
            "-f",
            GLOBAL_OPTS.config_path,
            buf,
            NULL
        };
        pid_t pid;
        int exit_status;
        if (posix_spawnp(&pid, "sysrc", NULL, NULL, args, environ) != 0) {
            warnx("failed to start sysrc(8)");
        } else if (waitpid(pid, &exit_status, 0) < 0) {
            warn("waitpid(2) for sysrc(8) failed");
        } else if (exit_status != 0) {
            warnx("sysrc(8) failed to set timezone");
        } else {
            LOG_VERBOSE("Set timezone to: \"%s\"\n", buf);
        }
        free(buf);
        request_restart();
    }
    if (state->up_down || state->back_down ||
        state->sel_down || state->down_down) {
        set_tz_screen_reset(screen, state);
        return true;
    }
    if (screen->need_redraw) {
        screen->need_redraw = false;
        lcd_display_control(state->lcd, LCD_CURSOR_NO_BLINK, LCD_CURSOR_OFF,
                            LCD_DISPLAY_ON);
        lcd_clear(state->lcd);
        lcd_move_to(state->lcd, 0, 0);
        lcd_write_string(state->lcd, "Done!");
    }
    return false;
}

static bool set_tz_screen_dispatch(SetTZScreen *screen,
                                   SensorState *state) {
    if (state->force_draw) {
        screen->need_redraw = true; 
    }
    SetTZScreenAction action;
    switch (screen->stage) {
    case SET_TZ_OFFSET:
        action = set_tz_dispatch_select_offset(&screen->off_sel_stage, state,
                                               &screen->offset, "Offset",
                                               screen->need_redraw, true);
        screen->need_redraw = false;
        switch (action) {
        case SET_TZ_BACK:
            set_tz_screen_reset(screen, state);
            return true;
        case SET_TZ_CONTINUE:
            break; // ignore
        case SET_TZ_NEXT:
            screen->need_redraw = true;
            ++screen->stage;
            break;
        }
        break;
    case SET_TZ_HAS_DST:
        set_tz_prompt_has_dst(screen, state);
        break;
    case SET_TZ_DST_OFFSET:
        action = set_tz_dispatch_select_offset(&screen->off_sel_stage, state,
                                               &screen->dst_offset,
                                               "DST Offset",
                                               screen->need_redraw, true);
        screen->need_redraw = false;
        switch (action) {
        case SET_TZ_BACK:
            --screen->stage;
            screen->need_redraw = true;
        case SET_TZ_CONTINUE:
            break; // ignore
        case SET_TZ_NEXT:
            screen->need_redraw = true;
            ++screen->stage;
            break;
        }
        break;
    case SET_TZ_DST_START:
        action = set_tz_dispatch_select_bound(screen, state,
                                              &screen->start_bound, "Start");
        switch (action) {
        case SET_TZ_BACK:
            --screen->stage;
            screen->need_redraw = true;
        case SET_TZ_CONTINUE:
            break; // ignore
        case SET_TZ_NEXT:
            screen->need_redraw = true;
            ++screen->stage;
            break;
        }
        break;
    case SET_TZ_DST_END:
        action = set_tz_dispatch_select_bound(screen, state,
                                              &screen->end_bound, "End");
        switch (action) {
        case SET_TZ_BACK:
            --screen->stage;
            screen->need_redraw = true;
        case SET_TZ_CONTINUE:
            break; // ignore
        case SET_TZ_NEXT:
            screen->need_redraw = true;
            ++screen->stage;
            break;
        }
        break;
    case SET_TZ_FINISH:
        return set_tz_screen_finish(screen, state);
    }
    return false;
}

SetTZScreen *set_tz_screen_new() {
    SetTZScreen *s = malloc_checked(sizeof(SetTZScreen));
    screen_init(&s->parent, "Set timezone",
                (ScreenDispatchFunc) set_tz_screen_dispatch,
                (ScreenCleanupFunc) free);
    set_tz_screen_reset(s, NULL);
    return s;
}
